import logging

from collections import namedtuple
from environs import Env

from wcraas_common import AMQPConfig


class RedisConfig(namedtuple("RedisConfig", ("host", "port"))):
    __slots__ = ()

    @classmethod
    def fromenv(cls):
        """
        Create a `wcraas_control.RedisConfig` from Environment Variables.

        >>> conf = RedisConfig.fromenv()
        >>> type(conf)
        <class 'config.RedisConfig'>
        >>> conf._fields
        ('host', 'port')
        >>> conf.host
        'localhost'
        >>> conf.port
        6379
        """
        env = Env()
        env.read_env()

        with env.prefixed("REDIS_"):
            return cls(host=env.str("HOST", "localhost"), port=env.int("PORT", 6379))


class Config(namedtuple("Config", ("amqp", "redis", "interval", "loglevel"))):
    __slots__ = ()

    @classmethod
    def fromenv(cls):
        """
        Create a `wcraas_control.Config` from Environment Variables.

        >>> conf = Config.fromenv()
        >>> type(conf)
        <class 'config.Config'>
        >>> conf._fields
        ('amqp', 'redis', 'interval', 'loglevel')
        >>> conf.amqp
        AMQPConfig(host='localhost', port=5672, user='guest', password='guest')
        >>> conf.redis
        RedisConfig(host='localhost', port=6379)
        >>> conf.interval
        10
        >>> conf.loglevel
        20
        """
        env = Env()
        env.read_env()

        return cls(
            amqp=AMQPConfig.fromenv(),
            redis=RedisConfig.fromenv(),
            interval=env.int("POLLING_INTERVAL", 10),
            loglevel=getattr(logging, env.str("LOGLEVEL", "INFO")),
        )



# -*- coding: utf-8 -*-

"""The WCraaS Control module is responsible for the orchestration of tasks in the platform."""

import asyncio
import logging
import traceback
import aio_pika

from collections import deque
from enum import unique, IntEnum

from aio_pika.patterns import RPC
from aioredis import create_redis_pool

from wcraas_common import AMQPConfig, WcraasWorker
from wcraas_control.config import RedisConfig


__all__ = ("RedisLockState", "ControlWorker")


@unique
class RedisLockState(IntEnum):
    FREE = 0
    LOCK = 1
    DONE = 2
    FAIL = 3


class ControlWorker(WcraasWorker):
    """
    Control Worker for the WCraaS platfrom, responsible for the orchestration of tasks.

    >>> from wcraas_control.config import Config
    >>> cn = ControlWorker(*Config.fromenv())
    """

    def __init__(
        self,
        amqp: AMQPConfig,
        redis: RedisConfig,
        interval: int,
        loglevel: int,
        *args,
        **kwargs,
    ):
        super().__init__(amqp, loglevel)
        self.redis = redis
        self.poll_interval = interval
        self.logger = logging.getLogger("wcraas.control")
        self.logger.setLevel(loglevel)

    async def crawl(self, url: str):
        """
        Given URL orchestrate crawling of the target.

        :param url: Entrypoint URL for crawling the target.
        :type url: string
        """
        redis_pool = await create_redis_pool(self.redis)
        urls = deque((url,))
        async with self._amqp_pool.acquire() as channel:  # type: aio_pika.Channel
            rpc = await RPC.create(channel)
            while True:
                # If there are no URLs left to crawl exit.
                if not urls:
                    break
                _url = urls.pop()
                self.logger.info(f"Processing {_url} ...")
                # Acquire lock for the URL
                await redis_pool.set(_url, RedisLockState.LOCK.value)
                # Delegate discovery through RPC
                try:
                    resp = await rpc.proxy.discover(url=_url)
                    self.logger.info(f"RPC for {_url} completed successfully!")
                    self.logger.debug(f"RPC response for {_url}:")
                    self.logger.debug(resp)
                except Exception as err:
                    self.logger.error(traceback.format_exc())
                    self.logger.error(err)
                    await redis_pool.set(_url, RedisLockState.FAIL.value)
                    continue
                await redis_pool.set(_url, RedisLockState.DONE.value)
                # For all inbound URLs (== same protocol, host, port) check if the URL
                # exists in redis & is not FREE; if so skip it otherwise add it to the
                # deque of URLs to crawl.
                for inbound_url in resp["data"]["inbound"]:
                    self.logger.info(f"Checking {inbound_url} against cache ...")
                    if await redis_pool.get(inbound_url):
                        self.logger.info(f"Skipping {inbound_url} due to cache hit!")
                        continue
                    urls.append(inbound_url)
                    self.logger.info(f"Adding {inbound_url} due to cache miss!")
                await asyncio.sleep(self.poll_interval)

    async def start(self): pass

    async def list_collections(self):
        """
        List the collections available at the storage node.
        """
        async with self._amqp_pool.acquire() as channel:
            rpc = await RPC.create(channel)
            resp = await rpc.proxy.list_collections()
        return resp

    def __repr__(self):
        return f"{self.__class__.__name__}(amqp={self.amqp}, redis={self.redis}, interval={self.poll_interval}, loglevel={self.loglevel})"



# -*- coding: utf-8 -*-

"""Console script for wcraas_control."""
import asyncio
import sys
import click

from wcraas_control import ControlWorker, Config


@click.command()
@click.argument("target")
def main(target):
    worker = ControlWorker(*Config.fromenv())
    asyncio.get_event_loop().run_until_complete(worker.crawl(target))


@click.command()
def list_collections():
    worker = ControlWorker(*Config.fromenv())
    resp = asyncio.get_event_loop().run_until_complete(worker.list_collections())
    print(resp)

if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'aio-pika',
    'aioredis',
    'Click>=7.0',
    'environs',
    'wcraas-common',
    'yarl>=1.3.0',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Kolokotronis Panagiotis",
    author_email='panagiks@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="WCraaS Storage Service",
    entry_points={
        'console_scripts': [
            'wcraas_control=wcraas_control.cli:main',
            'wcraas_list_collections=wcraas_control.cli:list_collections',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='wcraas_control',
    name='wcraas_control',
    packages=find_packages(include=['wcraas_control', 'wcraas_control.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/WCraaS/wcraas_control',
    version='0.1.2',
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""Top-level package for wcraas_control."""

__author__ = """Kolokotronis Panagiotis"""
__email__ = "panagiks@gmail.com"
__version__ = "__version__ = '0.1.2'"

from wcraas_control.config import Config  # NOQA
from wcraas_control.wcraas_control import ControlWorker  # NOQA
